import time
import websocket

def waitHandshake(ws, timeout=2):
    """
    Waits to receive exactly b'OK' as the vnc proxmox handshake.
    If it does not arrive before the timeout, returns False.
    """
    end_time = time.time() + timeout

    while time.time() < end_time:
        try:
            data = ws.recv()
        except websocket.WebSocketTimeoutException:
            #No handsake received yet,
            continue
        except Exception as e:
            print("Error while waiting for handshake: ", e)
            return False

        if data == b'OK':
            return True

    print("Timeout while waiting for handshake.")
    return False


import select
import time

import websocket

class TerminalHandler:
    def __init__(self, ws_connection):
        self.ws_connection = ws_connection

    def execCommand(self, command: str):
        for char in command:
            self.ws_connection.ws.send("0:1:" + char)
        self.ws_connection.ws.send("0:1:\n")  # Enter

    def sendInput(self, input_data: str):
        for char in input_data:
            self.ws_connection.ws.send("0:1:" + char)

    def sendBinaryInput(self, input_data: bytes):
        for byte in input_data:
            self.ws_connection.ws.send(f"0:1:{chr(byte)}")

    def readUntilPrompt(self, term_prompt="root@pve"):
        buffer = ""
        while True:
            data = self.ws_connection.ws.recv()
            if not data:
                break
            buffer += data.decode("utf-8", errors="ignore")
            if term_prompt in buffer:
                break
        return buffer

    def readTerm(self, wait_time=0.5):
        buffer = ""
        last_data_time = time.time()

        while True:
            rlist, _, _ = select.select([self.ws_connection.ws.sock], [], [], 0)
            if rlist:
                try:
                    data = self.ws_connection.ws.recv()
                    if data:
                        buffer += data.decode("utf-8", errors="ignore")
                        #print(data.decode("utf-8", errors="ignore")) print data
                        last_data_time = time.time()
                    else:
                        break
                except websocket.WebSocketConnectionClosedException:
                    print("Connection closed")
                    break
            else:
                if time.time() - last_data_time > wait_time:
                    break
                time.sleep(0.01)
        return buffer



from proxmoxer import ProxmoxAPI

class ProxmoxConnection:
    _instance = None

    @classmethod
    def get(cls, setup_data=None):
        # Primera inicialització
        if cls._instance is None:
            if setup_data is None:
                raise RuntimeError("setup_data és necessari la primera vegada")

            cls._instance = ProxmoxAPI(
                setup_data.hostname,
                user=setup_data.user,
                password=setup_data.password,
                otp=setup_data.otp,
                verify_ssl=False,
                port=8006,
            )

        return cls._instance



import threading
import time
import websocket
import ssl

class WSConnection:
    def __init__(self, ws_url, cookie_header=None, keep_alive_interval=10):
        self.ws_url = ws_url
        self.cookie_header = cookie_header
        self.keep_alive_interval = keep_alive_interval
        self.ws = None
        self._keep_alive_thread = None
        self._keep_alive_running = False

    def connect(self):
        self.ws = websocket.create_connection(
            self.ws_url,
            cookie=self.cookie_header,
            sslopt={"cert_reqs": ssl.CERT_NONE, "check_hostname": False}
        )
        

    def disconnect(self):
        self.stopKeepAlive()
        if self.ws:
            try:
                self.ws.close()
            except:
                pass
        self.ws = None

    # ------------------ Keep Alive ------------------
    def startKeepAlive(self):
        if self._keep_alive_thread is not None:
            return

        self._keep_alive_running = True

        def run():
            while self._keep_alive_running:
                try:
                    self.ws.send("2")
                except:
                    break
                time.sleep(self.keep_alive_interval)

        self._keep_alive_thread = threading.Thread(target=run, daemon=True)
        self._keep_alive_thread.start()

    def stopKeepAlive(self):
        self._keep_alive_running = False
        self._keep_alive_thread = None



import urllib.parse

from .utils import * 
from ..proxmoxer.core import ProxmoxAPI
from .connection import WSConnection
from .terminal import TerminalHandler

class ProxVNC:
    def __init__(self, api: "ProxmoxAPI" = None, *, url=None, node=None, user=None, shell_port=None, shell_ticket=None, pve_auth_cookie=None):
        self._store = {}
        if all(x is not None for x in [api]): # Provide only ProxmoxAPI (User, node default)
            self._store["api"] = api
            self._store["url"] = api._store["base_url"]

            if node is not None:
                if node not in [n["node"] for n in api.nodes.get()]:
                    raise ValueError(f"Node '{node}' not found in Proxmox cluster.")
                self._store["node"] = node
            else:
                self._store["node"] = api.nodes.get()[0]["node"] # ex: "pve" by default first node

            self._store["user"] = api.access.users.get()[0]["userid"] # ex: "root@pam"

        elif all(x is not None for x in [url, node, user, shell_port, shell_ticket, pve_auth_cookie]):
            for key, value in [("url", url), ("node", node), ("user", user), ("shell_port", shell_port), ("shell_ticket", shell_ticket), ("pve_auth_cookie", pve_auth_cookie)]:
                self._store[key] = value
                
        else:
            raise ValueError("Provide either ProxmoxAPI or all parameters.")

        self.connection = None
        self.terminal = None

    def connect(self, lxc=None, timeoutHandshake=2):
        api = self._store.get("api")
        wss_url = self._store["url"].replace("https://", "wss://", 1)

        if api:

            pve_cookie, _ = api.get_tokens()
            cookie_header = f"PVEAuthCookie={pve_cookie}"

            if lxc is not None:
                termproxy = api.nodes(self._store["node"]).lxc(lxc).termproxy.post()

                shell_ticket = termproxy["ticket"]
                shell_port = termproxy["port"]
                ws_url = f"{wss_url}/nodes/{self._store["node"]}/lxc/{lxc}/vncwebsocket?port={shell_port}&vncticket={urllib.parse.quote_plus(shell_ticket)}"

            else:
                # Connect to the node terminal
                termproxy = api.nodes(self._store["node"]).termproxy.post()

                shell_ticket = termproxy["ticket"]
                shell_port = termproxy["port"]
                ws_url = f"{wss_url}/nodes/{self._store["node"]}/vncwebsocket?port={shell_port}&vncticket={urllib.parse.quote_plus(shell_ticket)}"



            self.connection = WSConnection(ws_url, cookie_header)
            self.connection.connect()

            self.connection.ws.send(f"{self._store["user"]}:{shell_ticket}\n")
            self.connection.ws.send("1:86:24:")  # resolution

            if not waitHandshake(self.connection.ws, timeoutHandshake):
                raise TimeoutError("Handshake timed out.")
            
            self.connection.startKeepAlive()

            self.terminal = TerminalHandler(self.connection)
        else:

            if lxc is not None:
                ws_url = f"{wss_url}/api2/json/nodes/{self._store['node']}/lxc/{lxc}/vncwebsocket?port={self._store['shell_port']}&vncticket={urllib.parse.quote_plus(self._store['shell_ticket'])}"
            else:
                ws_url = f"{wss_url}/api2/json/nodes/{self._store["node"]}/vncwebsocket?port={self._store['shell_port']}&vncticket={urllib.parse.quote_plus(self._store['shell_ticket'])}"
            
            cookie_header=f"PVEAuthCookie={self._store['pve_auth_cookie']}"

            self.connection = WSConnection(ws_url, cookie_header)
            self.connection.connect()

            self.connection.ws.send(f"{self._store["user"]}:{self._store['shell_ticket']}\n")
            self.connection.ws.send("1:86:24:")  # resolution

            if not waitHandshake(self.connection.ws, timeoutHandshake):
                raise TimeoutError("Handshake timed out.")
            
            self.connection.startKeepAlive()

            self.terminal = TerminalHandler(self.connection)


# ---------------- Wrappers ----------------
    def execCommand(self, command: str):
        if self.terminal is None:
            raise RuntimeError("Terminal not initialized. Call connect() first.")
        self.terminal.execCommand(command)

    def readUntilPrompt(self, termPrompt="root@pve"):
        if self.terminal is None:
            raise RuntimeError("Terminal not initialized. Call connect() first.")
        return self.terminal.readUntilPrompt(termPrompt)

    def readTerm(self, waitTime=0.5):
        if self.terminal is None:
            raise RuntimeError("Terminal not initialized. Call connect() first.")
        return self.terminal.readTerm(waitTime)
    
    def disconnect(self):
        if self.terminal is not None:
            self.connection.disconnect()
            self.connection = None
            self.terminal = None

    def sendInput(self, input_data: str):
        if self.terminal is None:
            raise RuntimeError("Terminal not initialized. Call connect() first.")
        self.terminal.sendInput(input_data)
    
    def sendBinaryInput(self, input_data: bytes):
        if self.terminal is None:
            raise RuntimeError("Terminal not initialized. Call connect() first.")
        self.terminal.sendBinaryInput(input_data)